var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
import { isJust, nothing, just, isNothing, withDefault, map, map2, map3, map4, map5, mapN, andThen, caseOf, equals } from './maybe';
import * as chai from 'chai';
var expect = chai.expect;
describe('Maybe', function () {
    var add1 = function (n) { return n + 1; };
    describe('isJust', function () {
        it('should return false when Nothing is provided', function () {
            expect(isJust(nothing())).to.be.false;
        });
        it('should return true when Just is provided', function () {
            expect(isJust(just('hola'))).to.be.true;
        });
        it('should return false when null is provided', function () {
            expect(isJust(null)).to.be.false;
        });
        it('should return false when undefined is provided', function () {
            expect(isJust(undefined)).to.be.false;
        });
        it('should narrow the type', function () {
            var v = just('hello');
            if (isJust(v)) {
                expect(v._value).to.eq('hello');
            }
            else {
                throw new Error('Impossible state');
            }
        });
    });
    describe('isNothing', function () {
        it('should return false when Just is provided', function () {
            expect(isNothing(just('hola'))).to.be.false;
        });
        it('should return true when Nothing is provided', function () {
            expect(isNothing(nothing())).to.be.true;
        });
        it('should return true when null is provided', function () {
            expect(isNothing(null)).to.be.true;
        });
        it('should return true when undefined is provided', function () {
            expect(isNothing(undefined)).to.be.true;
        });
        it('should narrow the type', function () {
            var v = nothing();
            if (isNothing(v)) {
                expect(v).to.eql(nothing());
            }
            else {
                throw new Error('Impossible state');
            }
        });
    });
    describe('withDefault', function () {
        it('should return the Just value when Just is provided', function () {
            expect(withDefault(just(6), 0)).to.equal(6);
        });
        it('should return the default value when Nothing is provided', function () {
            expect(withDefault(nothing(), 0)).to.equal(0);
        });
        it('should return the default value when null is provided', function () {
            expect(withDefault(null, 1)).to.equal(1);
        });
        it('should return the default value when undefined is provided', function () {
            expect(withDefault(undefined, 2)).to.equal(2);
        });
    });
    describe('map', function () {
        it('should return Just when mapping over Just', function () {
            var maybeFour = just(4);
            var maybeFive = map(add1, maybeFour);
            expect(isJust(maybeFive)).to.be.true;
        });
        it('should return Nothing when mapping over Nothing', function () {
            var result = map(add1, nothing());
            expect(isNothing(result)).to.be.true;
        });
        it('should be isNothing === true when mapping over null', function () {
            var result = map(add1, null);
            expect(isNothing(result)).to.be.true;
        });
        it('should be isNothing === true when mapping over undefined', function () {
            var result = map(add1, undefined);
            expect(isNothing(result)).to.be.true;
        });
        it('should return 5 when adding 1 to Just(5)', function () {
            var maybeFour = just(4);
            var maybeFive = map(add1, maybeFour);
            expect(withDefault(maybeFive, 0)).to.equal(5);
        });
    });
    describe('map2', function () {
        var fullname = function (a, b) { return a + ' ' + b; };
        it('should return the fullname when appending Just("name") to Just("surname")', function () {
            var maybeName = just('John');
            var maybeSurame = just('Doe');
            var maybeFullname = map2(fullname, maybeName, maybeSurame);
            expect(withDefault(maybeFullname, '-')).to.equal('John Doe');
        });
        it('should return Nothing when appending Just() with Nothing', function () {
            var maybeName = just('John');
            var maybeFullname = map2(fullname, maybeName, nothing());
            expect(isNothing(maybeFullname)).to.be.true;
        });
        it('another example with numbers', function () {
            var safeParseInt = function (num) {
                var n = parseInt(num, 10);
                return isNaN(n) ? nothing() : just(n);
            };
            var sum = function (x, y) { return x + y; };
            var safeStringSum = function (x, y) {
                return map2(sum, safeParseInt(x), safeParseInt(y));
            };
            expect(safeStringSum('1', '2')).to.deep.equal(just(3));
            expect(safeStringSum('a', '2')).to.deep.equal(nothing());
        });
    });
    [map3, map4, map5].forEach(function (fn, index) {
        var n = index + 3;
        describe("map" + n, function () {
            var gt = function () {
                var nums = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    nums[_i] = arguments[_i];
                }
                return nums.every(function (num) { return num > 3; });
            };
            var just4Array = function (len) {
                return Array.apply(null, Array(len)).map(function () { return just(4); });
            };
            it('should return the true when all are greater than 3', function () {
                var result = fn.apply(void 0, __spreadArrays([gt], just4Array(n)));
                expect(withDefault(result, false)).to.be.true;
            });
            it('should be Nothing when any is Nothing', function () {
                var arr = just4Array(n);
                arr[2] = nothing();
                var result = fn.apply(void 0, __spreadArrays([gt], arr));
                expect(isNothing(result)).to.be.true;
            });
        });
    });
    describe('mapN', function () {
        var fullname = function (a, b, c) { return a + ' ' + b + ' ' + c; };
        it('should return the fullname when appending Just("name") to Just("middle") to Just("surname")', function () {
            var maybeName = just('John');
            var maybeMiddleName = just('D');
            var maybeSurame = just('Doe');
            var maybeFullname = mapN(fullname, maybeName, maybeMiddleName, maybeSurame);
            expect(withDefault(maybeFullname, '-')).to.equal('John D Doe');
        });
        it('should return Nothing when appending Just() with Nothing', function () {
            var maybeName = just('John');
            var maybeSurame = just('Doe');
            var maybeFullname = mapN(fullname, maybeName, nothing(), maybeSurame);
            expect(isNothing(maybeFullname)).to.be.true;
        });
    });
    describe('andThen', function () {
        var head = function (arr) { return (arr.length > 0 ? just(arr[0]) : nothing()); };
        it('should return Just(computation result) when Just is provided as argument', function () {
            var result = andThen(head, just(['a', 'b', 'c']));
            expect(withDefault(result, '')).to.equal('a');
        });
        it('should return Nothing when Nothing is provided as argument', function () {
            var result = andThen(head, nothing());
            expect(withDefault(result, '')).to.equal('');
        });
    });
    describe('caseOf', function () {
        it('should Launch 5 missiles', function () {
            var result = caseOf({
                Nothing: function () { return 'zzz'; },
                Just: function (n) { return "Launch " + n + " missiles"; }
            }, just('5'));
            expect(result).to.equal('Launch 5 missiles');
        });
        it('should zzz', function () {
            var result = caseOf({
                Nothing: function () { return 'zzz'; },
                Just: function (n) { return "Launch " + n + " missiles"; }
            }, nothing());
            expect(result).to.equal('zzz');
        });
    });
    describe('equals', function () {
        it('just(0) and just(0) should be equal', function () {
            expect(equals(just(0), just(0))).to.be.true;
        });
        it('just(5) and just(0) should not be equal', function () {
            expect(equals(just(5), just(0))).not.to.be.true;
        });
        it('nothing() and just(5) should not be equal', function () {
            expect(equals(nothing(), just(5))).not.to.be.true;
        });
        it('nothing() and nothing(5) should be equal', function () {
            expect(equals(nothing(), nothing())).to.be.true;
        });
        it('nothing() and null should be equal', function () {
            expect(equals(nothing(), null)).to.be.true;
        });
        it('nothing() and undefined should be equal', function () {
            expect(equals(nothing(), undefined)).to.be.true;
        });
    });
    describe('examples', function () {
        it('should work 1', function () {
            var user = {
                email: 'user@example.com',
                name: just('John'),
                surname: just('Doe')
            };
            var getFullName = function (name, surname) {
                return name + " " + surname;
            };
            var maybeFullname = map2(getFullName, user.name, user.surname);
            var fullName = withDefault(maybeFullname, '');
            expect(fullName).to.equal('John Doe');
        });
        it('should work 2', function () {
            var prices = [
                just(300),
                nothing(),
                just(500),
                just(150),
                nothing()
            ];
            var sum = function (a, b) { return a + b; };
            var total = withDefault(prices
                .filter(function (n) { return !equals(n, nothing()); })
                .reduce(function (acc, current) { return map2(sum, acc, current); }, just(0)), 0);
            expect(total).to.equal(950);
        });
    });
});
